@@ -62,6 +62,7 @@
 /*
  * private prototypes
  */
+static gboolean gok_spy_is_menu_role (AccessibleRole role);
 
 static void gok_spy_focus_listener (const AccessibleEvent *event,
                                     void *user_data);
@@ -78,12 +79,15 @@ SPIBoolean gok_spy_key_listener (const AccessibleKeystroke *key,
                                  void *user_data);
 
 AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
+				     GokSpySearchType type,
                                      AccessibleRole anAccessiblerole);
 
-static void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole Role);
+static void gok_spy_find_children (AccessibleNode* pNode, GokSpySearchType type, AccessibleRole Role);
 
 static gboolean gok_spy_has_child_aux (Accessible* accessible,
-                                       AccessibleRole role, gint children,
+				       GokSpySearchType search_type,
+                                       AccessibleRole role,
+				       gint children,
                                        gint depth, gint cdepth,
                                        gboolean* found);
 
@@ -114,7 +118,7 @@ static void* m_ptheMouseButtonListener;
 static Accessible* m_ptheWindowAccessible;
 
 /* pointer to the current accessible with editable text */
-static Accessible* m_ptheAccessibleWithEditableText;
+static Accessible* m_ptheAccessibleWithText;
 
 static AccessibleEventListener* focusListener;
 static AccessibleEventListener* defunctnessListener;
@@ -126,8 +130,7 @@ static gboolean m_gokSpyOpen = FALSE;
 static gint explicitrefs;
 static gint explicitunrefs;
 static gint implicitrefs;
-
-static guint spy_handler_id = 0;
+static gulong keyboardmods;
 
 /* 
  * for debugging
@@ -169,6 +172,7 @@ void gok_spy_accessible_implicit_ref(Accessible* accessible)
 	if (accessible != NULL)
 	{
 		implicitrefs++;
+#if defined(ENABLE_LOGGING_NORMAL)
 		s = Accessible_getName (accessible);
 		if (s == NULL)
 		{
@@ -179,6 +183,7 @@ void gok_spy_accessible_implicit_ref(Accessible* accessible)
 			gok_log ("%#x %s", accessible, s);
 			SPI_freeString (s);
 		}
+#endif
 	}
 	else
 	{
@@ -220,11 +225,10 @@ gboolean gok_spy_accessible_is_okay(Accessible* accessible)
 {
 	gboolean returncode = FALSE;
 	gok_log_enter();
-	if (accessible != NULL)
+	if (accessible != NULL) 
 	{
-		/*
+		/* FIXME: we should have a better way of doing this! */
 		returncode = cspi_accessible_is_a (accessible, "IDL:Accessibility/Accessible:1.0");
-		*/
 		/*
 		if (Accessible_isApplication(accessible) ||
 		    Accessible_isComponent(accessible))
@@ -242,11 +246,6 @@ gboolean gok_spy_accessible_is_okay(Accessible* accessible)
 				returncode = TRUE;
 			}
 			*/
-		/* kludge - check to see if there is even a window */
-		if (m_ptheWindowAccessible != NULL)
-		{
-			returncode = TRUE;
-		}
 	}
 	gok_log("returning: %d",returncode);
 	gok_log_leave();
@@ -275,7 +274,7 @@ void gok_spy_open()
 		m_ptheAppChangeListener = NULL;
 		m_ptheWindowChangeListener = NULL;
 		m_ptheWindowAccessible = NULL;
-		m_ptheAccessibleWithEditableText = NULL;
+		m_ptheAccessibleWithText = NULL;
 
 		focusListener = SPI_createAccessibleEventListener (
 			gok_spy_focus_listener, NULL);
@@ -292,8 +291,8 @@ void gok_spy_open()
 
 		success = SPI_registerGlobalEventListener (focusListener, "focus:"); 
 		gok_log ("SPI_registerGlobalEventListener for focus events returned: %d" , success);
-		/* success = SPI_registerGlobalEventListener (defunctnessListener, "object:state-changed:defunct");
-		gok_log ("SPI_registerGlobalEventListener for defunct state events returned: %d" , success); */
+		success = SPI_registerGlobalEventListener (defunctnessListener, "object:state-changed:defunct");
+		gok_log ("SPI_registerGlobalEventListener for defunct state events returned: %d" , success); 
 		success = SPI_registerDeviceEventListener (mouseListener,
 			SPI_BUTTON_PRESSED | SPI_BUTTON_RELEASED, NULL);
 		gok_log ("SPI_registerDeviceEventListener for mouse events returned: %d" , success);
@@ -310,6 +309,10 @@ void gok_spy_open()
 			windowDeactivateListener, "window:deactivate");
 		gok_log ("SPI_registerGlobalEventListener for window:deactivate returned: %d", success);
 
+		success = SPI_registerGlobalEventListener (
+			windowDeactivateListener, "window:destroy");
+		gok_log ("SPI_registerGlobalEventListener for window:destroy returned: %d", success);
+
 	}
 	gok_log_leave ();
 }
@@ -359,8 +362,8 @@ void gok_spy_close(void)
 		AccessibleEventListener_unref(modifierListener);
 
 		if (m_ptheWindowAccessible != NULL) gok_spy_accessible_unref(m_ptheWindowAccessible);
-		if (m_ptheAccessibleWithEditableText != NULL) gok_spy_accessible_unref(m_ptheAccessibleWithEditableText);
-		
+		if (m_ptheAccessibleWithText != NULL) gok_spy_accessible_unref(m_ptheAccessibleWithText);
+
 		gok_log("accessible reference balance is  [%d] (should be zero)",explicitrefs + implicitrefs - explicitunrefs);
 		gok_log("explicit accessible references   [%d]",explicitrefs);
 		gok_log("explicit accessible dereferences [%d]",explicitunrefs);
@@ -376,7 +379,7 @@ void gok_spy_close(void)
 			g_free(en);
 		}
 		g_queue_free(eventQueue);
-
+		eventQueue = NULL;
 	}
 	
 	gok_log_leave ();
@@ -449,23 +452,22 @@ void gok_spy_deregister_mousebuttonlistener(MouseButtonListener* callback)
 * gok_spy_get_list 
 * 
 * @paccessible: The parent accessible to the list
+* @type: The type of list to return, i.e. the search criteria
 * @role: The role of the list members
 *
 * User must call gok_spy_free when finished with this list.
 * 
 * Returns: pointer to the list or NULL
 **/ 
-AccessibleNode* gok_spy_get_list( Accessible* paccessible, AccessibleRole role )
+AccessibleNode* gok_spy_get_list( Accessible* paccessible, GokSpySearchType type, AccessibleRole role )
 {
 	AccessibleNode* plist = NULL;
 	gok_log_enter();
-	plist = gok_spy_create_list( paccessible, role );
+	plist = gok_spy_create_list( paccessible, type, role );
 	gok_log_leave();
 	return plist;
 }
 
-
-
 /** 
 * gok_spy_refresh 
 * 
@@ -503,8 +505,8 @@ void gok_spy_free (AccessibleNode* pNode)
 	{
 		pNodeToFree = pNode;
 		pNode = pNode->pnext;
-		
-		SPI_freeString (pNodeToFree->pname);
+		/* we gfree this since we've dup'd the SPI string */
+		g_free (pNodeToFree->pname);
 		gok_spy_accessible_unref (pNodeToFree->paccessible);
 		g_free (pNodeToFree);
 	}
@@ -512,13 +514,13 @@ void gok_spy_free (AccessibleNode* pNode)
 }
 
 /** 
-* gok_spy_get_accessibleWithEditableText
+* gok_spy_get_accessibleWithText
 *
 * accessor
 **/
-Accessible* gok_spy_get_accessibleWithEditableText()
+Accessible* gok_spy_get_accessibleWithText()
 {
-	return m_ptheAccessibleWithEditableText;
+	return m_ptheAccessibleWithText;
 }
 
 /** 
@@ -531,13 +533,14 @@ Accessible* gok_spy_get_accessibleWithEditableText()
 * 
 * Returns: true if a child exists, false otherwise. 
 **/ 
-gboolean gok_spy_has_child (Accessible* accessible, AccessibleRole role) 
+gboolean gok_spy_has_child (Accessible* accessible, GokSpySearchType type, AccessibleRole role) 
 {
 	gboolean bfound = FALSE;
 	gok_log_enter();
 	if (gok_spy_accessible_is_okay(accessible) == TRUE)
 	{
-		gok_spy_has_child_aux (accessible, role, GOKSPY_GUI_BREADTH, GOKSPY_GUI_DEPTH, 0, &bfound);
+		gok_spy_has_child_aux (accessible, type, role, 
+				       GOKSPY_GUI_BREADTH, GOKSPY_GUI_DEPTH, 0, &bfound);
 	}
 	else
 	{
@@ -553,13 +556,13 @@ gboolean gok_spy_has_child (Accessible* accessible, AccessibleRole role)
 *
 * this should be called during idle time. 
 **/ 
-void gok_spy_check_queues(void)
+gboolean gok_spy_check_queues(void)
 {
 	EventNode* en = NULL;
 	char *s;
 
 	gok_log_enter();
-	while (g_queue_is_empty(eventQueue) == FALSE)
+	while (eventQueue && g_queue_is_empty(eventQueue) == FALSE)
 	{
 		en = (EventNode*)g_queue_pop_head(eventQueue);
 
@@ -585,31 +588,29 @@ void gok_spy_check_queues(void)
 			gok_log("GOKSPY_FOCUS_EVENT");
 			gok_spy_process_accessible(en->paccessible);
 			break;
-		/*
-		 *  TODO: revisit this after CSUN
-		 *
+		case GOKSPY_WINDOW_ACTIVATE_EVENT:
+			gok_log("GOKSPY_WINDOW_ACTIVATE_EVENT");
+			/* FIXME !! the ref needs to happen in the original event loop */
+			if (en->paccessible != m_ptheWindowAccessible) {
+				gok_spy_accessible_unref (m_ptheWindowAccessible);
+				m_ptheWindowAccessible = en->paccessible;
+				gok_spy_accessible_ref (m_ptheWindowAccessible);
+				((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
+			}
+			break;
 		case GOKSPY_DEFUNCT_EVENT:
 			gok_log("GOKSPY_DEFUNCT_EVENT");
-			gok_spy_accessible_unref (m_ptheWindowAccessible);
+			/* note: currently only defunct events for the current application 
+				are pushed on the queue */
+			gok_spy_accessible_unref (m_ptheWindowAccessible); /* for our 
+				reference book-keeping */
 			m_ptheWindowAccessible = NULL;
 			((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			break;
-		case GOKSPY_WINDOW_ACTIVATE_EVENT:
-			gok_log("GOKSPY_WINDOW_ACTIVATE_EVENT");
-			gok_spy_accessible_unref (m_ptheWindowAccessible);
-			m_ptheWindowAccessible = en->paccessible;
-			gok_spy_accessible_ref (m_ptheWindowAccessible);
-			((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
-			break;
-		*/
 		case GOKSPY_WINDOW_DEACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_DEACTIVATE_EVENT");
 
-			/* TODO: Determine if this workaround is necessary
-			 *       or if at-spi should be modified to
-			 *       not send out of order activate/deactivate
-			 *       events
-			 *
+			/*
 			 * The test below is needed to workaround the case
 			 * that a window:deactivate event may arrive
 			 * after a window:activate has been emitted by
@@ -631,6 +632,7 @@ void gok_spy_check_queues(void)
 		g_free(en);
 	}
 	gok_log_leave();
+	return FALSE; /* only return true if we left events in the queue! */
 }
 
 /** 
@@ -673,7 +675,11 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
 	ass = Accessible_getStateSet(accessible);
 	if (ass != NULL)
 	{
-		if ( AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ) == TRUE)
+		AccessibleRole role = Accessible_getRole (accessible);
+		/* assumption: if this thing manages its own children, then it 
+		 * is probably not worth traversing.
+		 */
+		if ( AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ))
 		{
 			worthiness = FALSE;
 		}
@@ -687,7 +693,8 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
 /** 
  * gok_spy_has_child_aux:
  * @accessible: Pointer to the accessible.
- * @role: The role you are searching for.
+ * @type: The search criteria to apply.
+ * @role: The role you are searching for, if @type is GOK_SPY_SEARCH_ROLE.
  * @children: the breadth of the gui tree to search (for optimization).
  * @depth: how deep to look in the gui tree (for optimization).
  * @cdepth: recursive variable that keeps track of current depth
@@ -699,10 +706,14 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
  *
  * Returns: true if a child exists, false otherwise.
  */
-gboolean gok_spy_has_child_aux (Accessible* accessible, AccessibleRole role, gint children, gint depth, gint cdepth, gboolean* found) 
+gboolean gok_spy_has_child_aux (Accessible* accessible, 
+				GokSpySearchType type,
+				AccessibleRole role, 
+				gint children, gint depth, gint cdepth, gboolean* found) 
 { 
     Accessible* pAccessibleChild; 
-	int child_count;
+    AccessibleStateSet *states;
+    int child_count;
     int i; 
 
     if (accessible == NULL) return *found; 
@@ -727,30 +738,96 @@ gboolean gok_spy_has_child_aux (Accessible* accessible, AccessibleRole role, gin
 		gok_spy_accessible_implicit_ref(pAccessibleChild); 
 		if (pAccessibleChild != NULL) 
 		{ 
-			/* does this child have the role we're looking for? */ 
-		    if (Accessible_getRole (pAccessibleChild) == role) 
+		    /* does this child have the role we're looking for? */ 
+		    if (type == GOK_SPY_SEARCH_ROLE) 
 		    {
-		    	gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
-		    	gok_spy_accessible_unref(pAccessibleChild);
-		    	*found = TRUE;
-	    		return *found;
-		    } 
-		    else /* search through the children of this child */ 
-		    { 
-		    	if (*found == TRUE)
-		    	{
-			    	gok_spy_accessible_unref(pAccessibleChild);
-		    		return *found;
-		    	}
-		    	gok_spy_has_child_aux (pAccessibleChild, role, children, depth, cdepth, found);
-		    	gok_spy_accessible_unref(pAccessibleChild);
-		    } 
+			    if (Accessible_getRole (pAccessibleChild) == role) {
+				    gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
+				    gok_spy_accessible_unref(pAccessibleChild);
+				    *found = TRUE;
+				    return *found;
+			    }
+		    }
+		    else if (type == GOK_SPY_SEARCH_MENU) {
+			    AccessibleRole a_role = Accessible_getRole (pAccessibleChild);
+			    if ((a_role == role) || 
+				gok_spy_is_menu_role (a_role)) {
+				    gchar *name = Accessible_getName (pAccessibleChild);
+				    /* menu items must have either text or an image! */
+				    if (name && (strlen (name) > 0)) { /* todo: what about icon-menus? */
+					    gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
+					    gok_spy_accessible_unref(pAccessibleChild);
+					    SPI_freeString (name);
+					    *found = TRUE;					    
+					    return *found;
+				    }
+			    }
+		    }
+		    else if (type == GOK_SPY_SEARCH_EDITABLE_TEXT) {
+			    if (Accessible_isEditableText(pAccessibleChild)) {
+				    *found = TRUE;
+				    gok_spy_accessible_unref(pAccessibleChild);
+				    return *found;
+			    }
+		    }
+		    gok_spy_has_child_aux (pAccessibleChild, type, role, children, depth, cdepth, found);
+		    gok_spy_accessible_unref(pAccessibleChild);
 		} 
 	}
 	
 	return *found;
 } 
 
+static gboolean
+gok_spy_is_visible (Accessible *accessible)
+{
+	gboolean retval = FALSE;
+	if (accessible) {
+		AccessibleStateSet *states = Accessible_getStateSet (accessible);
+		if (AccessibleStateSet_contains (states, SPI_STATE_VISIBLE) &&
+		    AccessibleStateSet_contains (states, SPI_STATE_SHOWING))
+			retval = TRUE;
+		AccessibleStateSet_unref (states);
+	}
+	return retval;
+}
+
+static gboolean
+gok_spy_is_menu_role (AccessibleRole role)
+{
+	return ((role  ==  SPI_ROLE_MENU_ITEM) ||
+		(role ==  SPI_ROLE_CHECK_MENU_ITEM) ||
+		(role ==  SPI_ROLE_RADIO_MENU_ITEM) ||
+		(role ==  SPI_ROLE_MENU));
+}
+
+
+static gboolean
+gok_spy_is_ui (Accessible *accessible, AccessibleRole role)
+{
+	gboolean is_interesting_ui_component = FALSE;
+	switch (role) {
+	case SPI_ROLE_PUSH_BUTTON:
+	case SPI_ROLE_CHECK_BOX:
+	case SPI_ROLE_COMBO_BOX:
+	case SPI_ROLE_RADIO_BUTTON:
+	case SPI_ROLE_PAGE_TAB:
+	case SPI_ROLE_TOGGLE_BUTTON:
+		if (gok_spy_is_visible (accessible))
+			is_interesting_ui_component = TRUE;
+	default:
+		break;
+	}
+	/* EditableText and Hypertext components are always interesting */
+	if (!is_interesting_ui_component) {
+		if ((Accessible_isEditableText (accessible)) ||
+		    (Accessible_isHypertext (accessible)))
+			if (gok_spy_is_visible (accessible)) 
+				is_interesting_ui_component = TRUE;
+	}
+	return is_interesting_ui_component;
+}
+
 /**
  * gok_spy_find_children:
  * @pNode: Before execution of this function it is the node that you
@@ -765,9 +842,10 @@ gboolean gok_spy_has_child_aux (Accessible* accessible, AccessibleRole role, gin
  * beginning at pNode->pnext.  Please note that this function may also
  * modify pNode->paccessible.  This is a recursive function.
  */
-void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole role) 
+void gok_spy_find_children (AccessibleNode* pNode, GokSpySearchType type, AccessibleRole role) 
 { 
 	Accessible* pAccessibleChild;
+	AccessibleRole childrole;
 	Accessible* pAccessibleParent;
 	long child_count;
 	long i;
@@ -795,23 +873,36 @@ void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole role)
 
 		if (pAccessibleChild != NULL)
 		{
-			/* does this child have the role we're looking for? */
+			childrole = Accessible_getRole (pAccessibleChild);
 
-			/*
-			if (((role == SPI_ROLE_MENU_ITEM) && 
-			    ((Accessible_getRole (pAccessibleChild) ==  SPI_ROLE_MENU_ITEM) ||
-			     (Accessible_getRole (pAccessibleChild) ==  SPI_ROLE_CHECK_MENU_ITEM) ||
-			     (Accessible_getRole (pAccessibleChild) ==  SPI_ROLE_MENU))) ||
-			*/
-			if ((role == SPI_ROLE_MENU_ITEM)  /* always show a menu item */
-			    || (Accessible_getRole (pAccessibleChild) == role))
+			/* check role: special-casing some stuff REFACTOR THIS later TODO */
+			if ((type == GOK_SPY_SEARCH_ALL) /* include everything */
+			    ||
+			    /* include all menu-ish when search type: GOK_SPY_SEARCH_MENU*/
+			    ((type == GOK_SPY_SEARCH_MENU) && 
+			     gok_spy_is_menu_role (childrole))
+			    ||
+			    /* the search role matches the child role */
+			    ((type == GOK_SPY_SEARCH_ROLE) && (childrole == role))
+			    ||
+			    /* all ui */
+			    ((type == GOK_SPY_SEARCH_UI) && 
+			     (gok_spy_is_ui (pAccessibleChild, childrole)))
+			    ||
+				/* editable text */
+			    ((type == GOK_SPY_SEARCH_EDITABLE_TEXT) &&
+			     (Accessible_isEditableText(pAccessibleChild))))
 			{
 				gok_spy_append_node (pNode, pAccessibleChild);
+				if (childrole == SPI_ROLE_PAGE_TAB) {
+					pNode->paccessible = pAccessibleChild;
+					gok_spy_find_children (pNode, type, role);
+				}
 			}
 			else /* search through the children of this child */
 			{
 				pNode->paccessible = pAccessibleChild;
-				gok_spy_find_children (pNode, role);
+				gok_spy_find_children (pNode, type, role);
 			}
 
 		gok_spy_accessible_unref (pAccessibleChild);
@@ -831,6 +922,7 @@ void gok_spy_find_children (AccessibleNode* pNode, AccessibleRole role)
  * children found.
  */
 AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
+				     GokSpySearchType type,
                                      AccessibleRole anAccessibleRole)
 {
 	AccessibleNode nodeRoot;
@@ -843,12 +935,38 @@ AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
 	nodeRoot.pnext = NULL;
 	nodeRoot.pname = NULL;
 
-	gok_spy_find_children (&nodeRoot, anAccessibleRole);
+	gok_spy_find_children (&nodeRoot, type, anAccessibleRole);
 
 	gok_log_leave();
 	return nodeRoot.pnext;
 }
 
+static gboolean
+gok_spy_add_node (AccessibleNode *prev, Accessible *accessible, gint i, char *name)
+{
+	AccessibleNode *newnode = NULL;
+
+	if (!prev) return FALSE;
+
+	newnode = (AccessibleNode*)g_malloc(sizeof(AccessibleNode));
+	
+	if (newnode != NULL) {
+		newnode->paccessible = accessible;
+		gok_spy_accessible_ref (accessible);
+		newnode->is_link = (i >= 0) ? TRUE : FALSE;
+		newnode->link = MAX (0, i);
+		newnode->pname = name;
+		newnode->pnext = NULL;
+		gok_log("newnode->pname = %s", newnode->pname);
+		/* append this to end of list */
+		while (prev->pnext)
+			prev = prev->pnext;
+		prev->pnext = newnode;	
+		return TRUE;
+	}
+	return FALSE;
+}
+
 /**
  * gok_spy_append_node:
  * @pNode: An existing list of AccessibleNode to append to, must be
@@ -869,49 +987,162 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 {
 	AccessibleNode* pLastNode = NULL;
 	AccessibleNode* pNewNode = NULL;
-	char* pName = NULL;
+	AccessibleRelation** relations = NULL;
+	Accessible* targetAccessible = NULL;
+	AccessibleRelationType type = -1;
+	char* pName = NULL, *s;
+	int i = 0;
+	int j = 0;
+	int ntargets = 0;
 
 	g_assert (pNode != NULL);
 
 	gok_log_enter();
 
-	pName = Accessible_getName (pAccessible);
+        s = Accessible_getName (pAccessible);
+	if (s && strlen (s)) 
+	  pName = g_strdup (s);
+	if (s)
+	  SPI_freeString (s);
 
-	/* if the name is empty then ignore this child */ 
-	if ( (pName !=NULL) && (strlen (pName) != 0) )
-	{
-		/* create a new node */
-		pNewNode = (AccessibleNode*)g_malloc(sizeof(AccessibleNode));
-
-		if (pNewNode != NULL)
-		{
-			pNewNode->paccessible = pAccessible;
-			gok_spy_accessible_ref (pAccessible);
-
-			pNewNode->pname = pName;
-			gok_log("pNewNode->pname = %s",pNewNode->pname);
-			pNewNode->pnext = NULL;
-
-			pLastNode = pNode;
-			while (pLastNode->pnext != NULL)
-			{
-				pLastNode = pLastNode->pnext;
+	/* Hypertext objects are treated differently; one node per link */
+	if (Accessible_isHypertext (pAccessible)) {
+		AccessibleHypertext *hypertext = Accessible_getHypertext (pAccessible);
+		AccessibleText *text = Accessible_getText (pAccessible);
+		gint i, nlinks = AccessibleHypertext_getNLinks (hypertext);
+		for (i = 0; i < nlinks; ++i) {
+			AccessibleHyperlink *link = AccessibleHypertext_getLink (hypertext, i);
+			Accessible *link_object = AccessibleHyperlink_getObject (link, 0);
+			gchar *linkname;
+			long int start, end;
+			/* TODO: handle multi-anchor links! */
+			AccessibleHyperlink_getIndexRange (link, &start, &end);
+			linkname = AccessibleText_getText (text, start, end);
+			if (linkname) {
+				/* create and append a node */
+			  /* 
+			   * FIXME: when AccessibleHyperlink objects get un-broken,
+			   * we should pass the hyperlink object here instead of the 
+			   * hypertext object, so we can just invoke the default action
+			   * on the link!  But at the moment the link objects are not
+			   * actionable, so we have to fake it.
+			   */
+				gok_spy_add_node (pNode, pAccessible, i, g_strdup (linkname));
+				SPI_freeString (linkname);
 			}
-			pLastNode->pnext = pNewNode;
+			Accessible_unref (link_object);
 		}
-		else
+		AccessibleText_unref (text);
+		AccessibleHypertext_unref (hypertext);
+	}
+
+	/* if the name is empty then try LABELLED_BY, else ignore this child */ 
+	else {
+		/* look for a label */
+		if (pName == NULL)
 		{
-			SPI_freeString (pName);
+			gok_log("no name, so looking at relations...");
+			relations = Accessible_getRelationSet(pAccessible);
+			if (relations == NULL)	{
+				gok_log("no relationset for this object");
+				gok_log_leave();
+				return pNewNode;
+			}
+			while (relations[i]) {
+				type = AccessibleRelation_getRelationType(relations[i]);
+				if (type == SPI_RELATION_LABELED_BY) {
+					ntargets = AccessibleRelation_getNTargets(relations[i]);
+					for (j=0; j < ntargets; j++) {
+						targetAccessible = 
+							AccessibleRelation_getTarget(relations[i], j);
+						if (targetAccessible != NULL) {
+							s = Accessible_getName (targetAccessible);
+							
+							if (s != NULL) {
+								if ( strlen (s) != 0 )  {
+									pName = g_strdup (s);
+									SPI_freeString (s);
+									break; /* from for loop*/
+								}
+								else {
+									SPI_freeString(s);
+								}
+							}
+						}
+					}
+					break; /* from while loop */
+				}
+			}
+		}
+		
+		
+		if (pName != NULL) {
+			/* create a new node */
+			gok_spy_add_node (pNode, pAccessible, -1, pName);
+		}
+		else if (Accessible_isText (pAccessible)) {
+			/* 
+			* sadly, many text objects are still nameless - yet we want 
+			* to allow the user to get to them.
+			* first tack is to get the starting text... if the object is
+			* empty, then we can still "create" a name for it, though we 
+			* can't assign it a unique meaningful name.
+			*/
+			AccessibleText *text = Accessible_getText (pAccessible);
+			gint j = 0, len;
+			gchar *word, *gs = NULL;
+			if (text) {
+				long int start, end = 0;
+				do {
+					gchar *sp = gs;
+					word = AccessibleText_getTextAtOffset (text, end, 
+									       SPI_TEXT_BOUNDARY_WORD_END,
+									       &start, &end);
+					if (word) {
+						len = strlen (word);
+					}
+					else {
+						len = 0;
+					}
+					if (len) {
+						if (sp) {
+							gs = g_strconcat (sp, word, NULL);
+							g_free (sp);
+						}
+						else {
+							gs = g_strdup (word);
+						}
+					}
+					if (word) { 
+						SPI_freeString (word);
+					}
+					++j;
+				} while ((j < 3) && len);
+				AccessibleText_unref (text);
+			}
+			if (gs && strlen (gs)) {
+				pName = g_strconcat (gs, "...", NULL);
+			}
+			else {
+				if (gs) {
+					g_free (gs);
+				}
+				if (Accessible_isEditableText (pAccessible)) {
+					pName = g_strdup ("Text Entry (empty)");
+				}
+			}
+			if (pName != NULL) {
+				gok_spy_add_node (pNode, pAccessible, -1, pName);
+			}
 		}
 	}
-	else if (pName != NULL)
-	{
-		SPI_freeString (pName);
-	}
+	
+	while (pNode->pnext) 
+		pNode = pNode->pnext;
 
 	gok_log_leave();
 
-	return pNewNode;
+	return pNode;
 }
 
 /** 
@@ -928,10 +1159,10 @@ void gok_spy_focus_listener (const AccessibleEvent *event, void *user_data)
 	en = (EventNode*)g_malloc(sizeof(EventNode));
 	en->paccessible = event->source;
 	en->type = GOKSPY_FOCUS_EVENT;
-    g_queue_push_tail( eventQueue, en );
+	g_queue_push_tail( eventQueue, en );
 
 	gok_spy_add_idle_handler ();
-    gok_log_leave();
+	gok_log_leave();
 }
 
 /** 
@@ -944,17 +1175,16 @@ void gok_spy_defunctness_listener (const AccessibleEvent *event, void *user_data
 	EventNode* en; 
 	gok_log_enter();
 
-    if (event->source == m_ptheWindowAccessible)
-    {
-		en = (EventNode*)g_malloc(sizeof(EventNode));
-		en->paccessible = NULL;/*event->source;*/
+	if (event->source == m_ptheWindowAccessible) {
+		en = (EventNode*) g_malloc(sizeof(EventNode));
+		en->paccessible = NULL; /* the source accessible is defunct! */
 		en->type = GOKSPY_DEFUNCT_EVENT;
 		g_queue_push_tail ( eventQueue, en );
 		gok_spy_add_idle_handler ();
-    	gok_log_x("our window is gone!");
-    }
-
-    gok_log_leave();
+		gok_log_x("our window is gone!");
+	}
+	
+	gok_log_leave();
 }
 
 static void
@@ -1001,47 +1231,45 @@ gok_spy_window_deactivate_listener(const AccessibleEvent *event,
 **/ 
 void gok_spy_process_accessible (Accessible* accessible) 
 { 
-    Accessible* pnewWindowAccessible;
-    Accessible* ptempAccessible;
-
-    gok_log_enter();
-    g_assert(accessible != NULL);
-
-    pnewWindowAccessible = NULL;
+	Accessible* pnewWindowAccessible;
+	Accessible* ptempAccessible;
+	
+	gok_log_enter();
+	g_assert(accessible != NULL);
+	
+	pnewWindowAccessible = NULL;
 	ptempAccessible = NULL;
 	
 	/* start by assuming we want to keep this accessible around */
 	gok_spy_accessible_ref(accessible); 
 
 	/*
-	 *  check to see if focus is on an editable text object 
+	 *  check to see if focus is on a text object 
 	 */
-	gok_log("Checking for editable text...");
-	gok_log("m_ptheAccessibleWithEditableText [%d]",m_ptheAccessibleWithEditableText);
-	gok_log("                      accessible [%d]",accessible);
-	if (Accessible_isEditableText(accessible) == TRUE)
+	if (Accessible_isText(accessible) == TRUE)
 	{
-		gok_log("this thing has an editable text interface");
-		if (accessible != m_ptheAccessibleWithEditableText)
+		gok_log("this thing has a text interface");
+		if (accessible != m_ptheAccessibleWithText)
 		{
 			gok_spy_accessible_ref(accessible); 
-			gok_spy_accessible_unref(m_ptheAccessibleWithEditableText);
-			m_ptheAccessibleWithEditableText = accessible;
-			gok_log("found a (new) editable text interface"); 
+			gok_spy_accessible_unref(m_ptheAccessibleWithText);
+			m_ptheAccessibleWithText = accessible;
+			gok_log("found a (new) text interface"); 
 		}
 	}
 	else
 	{
-		gok_log("no editable text interface on this thing");
+		gok_log("no text interface on this thing");
 		/* we might not want to do this - needs user testing */
-		if (m_ptheAccessibleWithEditableText != NULL)
+		if (m_ptheAccessibleWithText != NULL)
 		{
-			gok_spy_accessible_unref(m_ptheAccessibleWithEditableText);
+			gok_spy_accessible_unref(m_ptheAccessibleWithText);
 		}
-		m_ptheAccessibleWithEditableText = NULL;
+		m_ptheAccessibleWithText = NULL;
 	}
 
-	/* bug with gedit toolbar has to do with needing to carry on up the parent chain 
+	/* 
+	   bug with gedit toolbar has to do with needing to carry on up the parent chain 
 	   if we do this we might lose context - need to balance pros and cons...
 	*/	
 
@@ -1053,34 +1281,32 @@ void gok_spy_process_accessible (Accessible* accessible)
 	gok_log("accessible [%d]",accessible);
 	pnewWindowAccessible = accessible;
 	while (gok_spy_check_window(Accessible_getRole(pnewWindowAccessible)) != TRUE)
-	 /*|| (ptempAccessible != NULL))*/
-	
 	{
 		gok_log("*");
-    	ptempAccessible = Accessible_getParent(pnewWindowAccessible);
-	    gok_spy_accessible_implicit_ref(ptempAccessible);
-	    if (ptempAccessible == NULL)
-	    {
-	    	break;
-	    } 
-    	gok_spy_accessible_unref(pnewWindowAccessible);
-   	 pnewWindowAccessible = ptempAccessible;
-    }
-	 
+		ptempAccessible = Accessible_getParent(pnewWindowAccessible);
+		gok_spy_accessible_implicit_ref(ptempAccessible);
+		if (ptempAccessible == NULL)
+		{
+			break;
+		} 
+		gok_spy_accessible_unref(pnewWindowAccessible);
+		pnewWindowAccessible = ptempAccessible;
+	}
+	
 	/*
      * check to see if we have a new window 
 	 */
 	gok_log("Comparing this container with existing...");
-    if (pnewWindowAccessible != m_ptheWindowAccessible)
-    {
-    	gok_log("different.");
+	if (pnewWindowAccessible != m_ptheWindowAccessible)
+	{
+		gok_log("different.");
 		gok_spy_accessible_unref (m_ptheWindowAccessible);
-    	m_ptheWindowAccessible = pnewWindowAccessible;
+		m_ptheWindowAccessible = pnewWindowAccessible;
 		((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible); 
 	}
 	else
 	{
-    	gok_log("same.");
+		gok_log("same.");
 		/* we have one ref too many (this isn't a different window) */
 		gok_spy_accessible_unref(pnewWindowAccessible);
 	}
@@ -1157,6 +1383,15 @@ gok_spy_mouse_listener (const AccessibleDeviceEvent *event, void *user_data)
         return is_switch_trigger;
 }
 
+/**
+ * gok_spy_get_modmask:
+ **/
+gulong gok_spy_get_modmask (void)
+{
+	return keyboardmods;
+}
+
+
 /** 
  * gok_spy_modifier_listener:
  * @event: the keyboard modifier event.
@@ -1167,10 +1402,13 @@ gok_spy_mouse_listener (const AccessibleDeviceEvent *event, void *user_data)
 void gok_spy_modifier_listener(const AccessibleEvent *event, void *user_data)
 {
 	const struct gok_modifier_keymask *m;
+       
+	keyboardmods = event->detail2;
 
 	gok_log_enter();
 	gok_log ("event->detail2 = %d", event->detail2);
 
+	
 	for (m = gok_modifier_keymasks; m->name != NULL; m++)
 	{
 		if (event->detail2 & m->keymask)
@@ -1191,9 +1429,7 @@ void gok_spy_modifier_listener(const AccessibleEvent *event, void *user_data)
 static gboolean
 gok_spy_idle_handler (gpointer data)
 {
-	spy_handler_id = 0;
-	gok_spy_check_queues ();
-	return FALSE;
+	return gok_spy_check_queues ();
 }
 
 /** 
@@ -1203,24 +1439,9 @@ gok_spy_idle_handler (gpointer data)
  */
 void gok_spy_add_idle_handler (void)
 {
-	if (spy_handler_id == 0) {
-		spy_handler_id = g_idle_add (gok_spy_idle_handler, NULL);
-	}
+	g_idle_add (gok_spy_idle_handler, NULL);
 }
 
-/** 
- * gok_spy_remove_idle_handler:
- *
- * Removes the idle handler to deal with spy events
- */
-void gok_spy_remove_idle_handler (void)
-{
-	if (spy_handler_id != 0) {
-		gtk_idle_remove (spy_handler_id);
-	}    
-}
-
-
 /*
 gboolean gok_spy_is_desktop(Accessible* pAccessible)
 {
@@ -1345,4 +1566,3 @@ AccessibleNode* gok_spy_get_applications( )
 	return plist;
 }
 */
-
